#include <cmath>
#include <fstream>
#include <stdlib.h>
#include "basics.h"
#include "shape.h"

using namespace std;

/*   Contents:
Shape element class methods
Shape class methods
Shape construction functions
Shape combination functions


 -------------  The shape element class -----------------

A shape element is a point, a line or an arc.
For an arc: kind=se_curve, x,y=centre, r=radius, arg1-arg2 is the range of normal angles
For a point: kind=se_point, x,y=position, r=0,  arg1-arg2   "     "         "
For a line segment: kind=se_line, arg1=arg2=normal angle, x,y,r are copied from the adjoining elements

A shape which uses these shape elements will have them linked in a cycle and have a pointer to one of them.
*/

int n_shape_elmts=0;

float opp_arg(float a) // argument of the opposite direction
{
  if (a>0) return a-PI;
  else return a+PI;
}

char arg_cmp(float a1,float a2)
{
  if (argeq(a1+PI,a2)) return 0;
  if (a2==PI) return 1;
  if (a1<=0) return (a1+PI>a2)*2-1; // should this be < or <= ?
  else return (a1-PI>a2)*2-1;
}

shape_elmt::shape_elmt()  { n_shape_elmts++; }

shape_elmt::~shape_elmt() { n_shape_elmts--; }

bool shape_elmt::on_line(int x1,int y1)  // is the point on the line?
{
  if (cos(arg1)*(x1-x)+sin(arg1)*(y1-y)-r<-2.5) return false;  // previously -1.5
  if (sin(arg1)*(x1-prev->x)-cos(arg1)*(y1-prev->y)>0.5) return false;
  if (-sin(arg1)*(x1-next->x)+cos(arg1)*(y1-next->y)>0.5) return false;
 return true;
}

bool shape_elmt::in_line(int x1,int y1) // is the point on the inside of the line?
{  
  return cos(arg1)*(x1-x)+sin(arg1)*(y1-y)<r-0.001; 
}

float shape_elmt::on_curve(int x1,int y1) // is the point on the curve? 
                                // If yes, returns arg. If no, returns 22.
{
  int x2,y2; float a;
  x2=x1-x; y2=y1-y;
  a=atan2((float)y2,(float)x2);
  if (arg1>arg2)
    if (a>arg1 || a<arg2) return a; else return 22;
  else 
    if (a>arg1 && a<arg2) return a; else return 22;
}

bool shape_elmt::in_curve(int x1,int y1) // is the point inside the curve?
{ return sq(x1-x)+sq(y1-y)<r*r; }

bool shape_elmt::in_chord(int x1,int y1) // is the point on the inside of the chord?
{ 
  float a; // between a1 & a2
  if (arg1>arg2) a=(arg1+arg2)/2+PI;
  else a=(arg1+arg2)/2;
  return (x1-x)*cos(a)+(y1-y)*sin(a)<r*cos(arg2-a);
}

void shape_elmt::invert()
{
  x=-x; y=-y;
  arg1=opp_arg(arg1);
  arg2=opp_arg(arg2);
}

void shape_elmt::describe()
{
  if (kind==se_point)
    orep << "Point (" << x << "," << y << "), args " << arg1 << " to " << arg2 << endl;
  if (kind==se_curve)
    orep << "Arc with centre (" << x << "," << y << ") and radius " << r << ", args " << arg1 << " to " << arg2 << endl;
  if (kind==se_line)
    orep << "Line segment (" << x << "," << y << " |" << r << "), arg " << arg1 << endl;
}

/* ------------- Shape class methods---------------- */

shape::shape(int k)
  {
    strokes=NULL; a1=a2=NULL;
    n_shapes++;
    r=lr=ud=0;
    max_x=max_y=99999;
    min_y=min_x=-99999;
    kind=k;
  }

shape::~shape()
  {
    shape_elmt *se,*se1;
    n_shapes--;

    if (strokes)
    {
      for(se=strokes;se;se=se1)
      {
        se1=se->next; 
        if (se1==strokes) se1=NULL;
        delete se;
      }
    }
    if (kind==sh_twoshapes)
    {
      delete a1;
      delete a2;
    }
    if (kind==sh_translation)
      delete a1;
  }

  void shape::calc_maxima()
  {
    if (kind>sh_hlin) return;
    max_x=lr+r;
    min_x=-max_x;
    max_y=ud+r;
    min_y=-max_y; // This only works for roundrects
  }

  shape_elmt* shape::find_se(float arg,bool eq) // do we need eq?
  {
    shape_elmt *se;
    se=strokes;
    while (!in_order(se->arg1,arg,se->arg2))
      se=se->next;
    if (eq && argeq(se->prev->arg1,arg)) return se->prev;
    return se;
  }

  void shape::add_se(int x,int y,int r,float arg1,float arg2)
  {
    shape_elmt *se=new shape_elmt;
    if (argeq(arg1,arg2)) se->kind=se_line;
    else if (r) se->kind=se_curve;
    else se->kind=se_point;
    se->x=x; se->y=y; se->r=r; se->arg1=arg1; se->arg2=arg2;
//    orep << "Added " << (int)se->kind Then x Then y Then r Then arg1 Then arg2 << endl;
    if (strokes)
    {
      if (se->kind==se_line && strokes->prev->kind==se_line)
      {
        delete se;
        return;
      }
      se->next=strokes;
      se->prev=strokes->prev;
      strokes->prev=se;
      se->prev->next=se;
    }
    else strokes=se->next=se->prev=se;

    if (in_order(arg1,0,arg2)) max_x=x+r;
    if (in_order(arg1,PI,arg2)) min_x=x-r;
    if (in_order(arg1,PI/2,arg2)) max_y=y+r;
    if (in_order(arg1,-PI/2,arg2)) min_y=y-r;
    n_strokes++;
  }

  void shape::make_strokes()
  {
    add_se(lr,-ud,r,-PI/2,0);
    add_se(lr,-ud,r,0,0);
    add_se(lr,ud,r,0,PI/2);
    add_se(lr,ud,r,PI/2,PI/2);
    add_se(-lr,ud,r,PI/2,PI);
    add_se(-lr,ud,r,PI,PI);
    add_se(-lr,-ud,r,PI,-PI/2);
    add_se(-lr,-ud,r,-PI/2,-PI/2);
  }

  void shape::describe()
  {
    shape_elmt *se;
    if (kind==sh_twoshapes)
    {
      a1->describe();
      orep << "        AND\n";
      a2->describe();
    }
    for(se=strokes;se;se=se->next)
    {
      se->describe();
      if (se->next==strokes) break;
    }
    orep << "Bounding rectangle: (" << min_x << "," << min_y << ")-(" << max_x << "," << max_y << ")\n";
  }

  bool shape::point_in(int x,int y)
  {
    switch (kind)
    {
    case sh_noshape: return false;

    case sh_point: return (x==0 && y==0);

    case sh_circl: return (x*x + y*y < r*r);

    case sh_rectangle: return (Abs(x)<lr && Abs(y)<ud);
 
    case sh_hlin: return (Abs(x)<lr && !y);

    case sh_roundrect:
      int x1,y1;

      x1=Abs(x);y1=Abs(y);

      if (x1>=lr+r) return false;
      if (y1>=ud+r) return false;
      if (x1<lr || y1<ud) return true;
      return (sq(x1-lr)+sq(y1-ud)<r*r);

    case sh_polystroke:
      shape_elmt *se;
      for (se=strokes;se;se=se->next)
      {
        if (se->kind==se_line)
          if (!se->in_line(x,y)) return false;
        if (se->next==strokes) break;
      }
      for (se=strokes;se;se=se->next)
      {
        if (se->kind==se_curve)
          if (se->in_curve(x,y)) return true;
        if (se->next==strokes) break;
      }
      for (se=strokes;se;se=se->next)
      {
        if (se->kind==se_curve)
          if (!se->in_chord(x,y)) return false;
        if (se->next==strokes) break;
      }
      return true;

    case sh_twoshapes:
      return a1->point_in(x,y) || a2->point_in(x,y);

    case sh_translation:
      return a1->point_in(x-lr, y-ud);

    }
    return false;
  }

  float shape::argument(int x,int y)
  {
    shape_elmt *se; float a; int x1,y1;

    switch (kind)
    {
    case sh_point: return atan2((float)y,(float)x);

    case sh_circl: return atan2((float)y,(float)x);

    case sh_rectangle:
      if (Abs(x)-lr>Abs(y)-ud)
        if (x>0) return false;
        else return PI;
      else
        if (y>0) return PI/2;
        else   return -PI/2;

    case sh_roundrect:
      x1=Abs(x)-lr;y1=Abs(y)-ud;
      if (x1>=0 && y1>=0)
        if (x>0)
          if (y>0)
            return atan2((float)y1,(float)x1);
          else return atan2((float)-y1,(float)x1);
        else
          if (y>0)
            return atan2((float)y1,(float)-x1);
          else return atan2((float)-y1,(float)-x1);
      else
        if (x1>y1)
          if (x>0) return false;
          else return PI;
        else
          if (y>0) return PI/2;
          else return -PI/2;

    case sh_hlin: return PI/2;


    case sh_polystroke:
      for (se=strokes;se;se=se->next)
      {
        if (se->kind==se_line)
          if (se->on_line(x,y))
          {
            return se->arg1;
          }
        if (se->next==strokes) break;
      }
      for (se=strokes;se;se=se->next)
      {
        if (se->kind==se_curve)
        {
          a=se->on_curve(x,y);
          if (a<20) return a;
        }
        if (se->next==strokes) break;
      }
      orep << "Couldn't find the argument of (" << x << "," << y << ").\n";
      describe();
      exit(0);

    case sh_twoshapes:
      if (a1->point_in(x,y)) return a1->argument(x,y);
      else                   return a2->argument(x,y);

    case sh_translation:
      return a1->argument(x-lr, y-ud);
    }
    return 1.23; // should be something more useful

  }

shape* shape::copy()
{
  shape* r=new shape(kind);
  *r=*this;

  r->strokes=NULL;
  r->n_strokes=0;
  if (kind==sh_polystroke)
  {
    shape_elmt* se=strokes;
    do {
      r->add_se(se->x,se->y,se->r,se->arg1,se->arg2);
      se=se->next;
    } while (se!=strokes);
  }
  else if (kind==sh_twoshapes)
  {
    r->a1=a1->copy();
    r->a2=a2->copy();
  }
  else if (kind==sh_translation)
    r->a1=a1->copy();
  return r;
}

shape* shape::copy_opp()
{
  shape *r; shape_elmt *se;
  r=new shape(kind);

  *r=*this;
  r->strokes=NULL;
  r->n_strokes=0;

  if (kind==sh_polystroke)
  {
    se=strokes;
    do
    {
      r->add_se(-se->x,-se->y,se->r,opp_arg(se->arg1),opp_arg(se->arg2));
      se=se->next;
    }
    while (se!=strokes);
  }
  else if (kind==sh_twoshapes)
  {
    r->a1=a1->copy_opp();
    r->a2=a2->copy_opp();
  }
  r->a1=a1->copy_opp();
  r->max_x=-min_x;
  r->min_x=-max_x;
  r->max_y=-min_y;
  r->min_y=-max_y;
  return r;
}


/* ------------ Shape construction functions ------------- */

shape *new_point()
{
  return new shape(sh_point);
}

shape *new_circl(int r)
{
  shape *c=new shape(sh_circl);
  c->r=r;
  c->calc_maxima();
  return c;
}

shape *new_rectangle(int lr,int ud)
{
  shape* r=new shape(sh_rectangle);
  r->lr=lr;
  r->ud=ud;
  r->calc_maxima();
  return r;
}

shape *new_roundrect(int lr,int ud,int ra)
{
  shape* r=new shape(sh_roundrect);
  r->lr=lr;
  r->ud=ud;
  r->r=ra;
  r->calc_maxima();
  return r;
}

shape *new_hlin(int lr)
{
  shape* h=new shape(sh_hlin);
  h->lr=lr;
  h->calc_maxima();
  return h;
}

shape *new_twoshapes(shape *s1,shape *s2)
{
  shape* t=new shape(sh_twoshapes);
  t->a1=s1;
  t->a2=s2;
  t->max_x=max2(s1->max_x,s2->max_x);
  t->min_x=min2(s1->min_x,s2->min_x);
  t->max_y=max2(s1->max_y,s2->max_y);
  t->min_y=min2(s1->min_y,s2->min_y);
  return t;
}


float sh_sidearg[10];

shape *make_polygon(int n,int *v)
{
  shape *p;
  int j,x1,x2,y1,y2;


  p=new shape(sh_polystroke);
  for(j=0;j<n;j++)
  {
    x1=v[j*2];
    y1=v[j*2+1];
    x2=v[(j*2+2)%(2*n)];
    y2=v[(j*2+3)%(2*n)];
    sh_sidearg[j]=atan2((float)x1-x2,(float)y2-y1);
  }
  for(j=0;j<n;j++)
  {
    p->add_se(v[j*2],v[j*2+1],0,sh_sidearg[(j-1+n)%n],sh_sidearg[j]);
    p->add_se(v[j*2],v[j*2+1],0,sh_sidearg[j],sh_sidearg[j]);
  }
  return p;
}

shape *translate(shape *s, int x, int y)
{
  shape* r=new shape(sh_translation);
  r->lr=x; r->ud=y; r->a1=s;
  r->max_x=x+s->max_x; r->min_x=x-s->min_x;
  r->max_y=y+s->max_y; r->min_y=y-s->min_y;
  return r;
}


/* -------------- Shape combination functions --------------------- */

#define IF if (0) orep

shape *combine_polystrokes(shape *s1,shape *s2)
{
  shape *r;
  shape_elmt *e1,*e2=NULL, *s1f0;
  char cmp;
  float arg2,ut;

  if (!s1 || !s2) { return new shape(0); }
  if (!s1->strokes) s1->make_strokes();
  if (!s2->strokes) s2->make_strokes();
  r=new shape(sh_polystroke);

 IF << "Combining " << (int)s1->kind << " & " << (int)s2->kind << endl;

  e1=s1f0=s1->find_se(0,true);
  ut=0;

  do
  {
    if (e1->kind==se_line)
    {
      e2=s2->find_se(opp_arg(e1->arg1),true);
      IF << "e1: ";// e1->describe();
      IF << "e2: ";// e2->describe();
      r->add_se(e1->x-e2->x,e1->y-e2->y,e1->r+e2->r,e1->arg1,e1->arg2);
      if (argeq(opp_arg(e2->arg2),e1->arg1)) { e2=e2->next; IF << "e2 NEXT" << endl; }
      ut=e1->arg2;
    }
    else
    {
      if (ut==0)
      {
        ut=e1->arg1;
        e2=s2->find_se(opp_arg(e1->arg1),true); // false
      }
      do
      {
        if (e2->kind==se_line) cmp=1;
        else
        {
          cmp=arg_cmp(e1->arg2,e2->arg2);
          IF << "Cmp(" << e1->arg2 Then e2->arg2 << ") = " << (int)cmp << endl;
        }
        if (cmp<0) arg2=e1->arg2;
        else arg2=opp_arg(e2->arg2);

        IF << "e1: ";// e1->describe();
        IF << "e2: ";// e2->describe();
        r->add_se(e1->x-e2->x,e1->y-e2->y,e1->r+e2->r,ut,arg2);

        if (cmp>=0) {e2=e2->next; IF << "e2 NEXT" << endl; }
        ut=arg2;
      }
      while (cmp>0);
    }
    e1=e1->next;
    IF << "e1 NEXT" << endl;
  }
  while (e1!=s1f0);
  IF << "finished shape" << endl;
  return r;
}


//   combines shapes s1 and s2
//The result (r) is such that if s1 and s2 overlap, (x2-x1,y2-y1) is inside shape r */

shape *combine_sh(shape *s1,shape *s2)
{
  if (!s1 || !s2) return new shape(sh_noshape);

  shape *r;
  shape_elmt *se;

  int k1=s1->kind;
  int k2=s2->kind;
  if (!k1 || !k2) return new shape(sh_noshape);

  if (k1==sh_twoshapes)
    return new_twoshapes(combine_sh(s1->a1,s2),combine_sh(s1->a2,s2));

  if (k2==sh_twoshapes)
    return new_twoshapes(combine_sh(s1,s2->a1),combine_sh(s1,s2->a2));

  if (k1==sh_point) return s2->copy_opp();
  if (k2==sh_point) return s1->copy();

  if (k1==sh_polystroke && k2==sh_circl)
  {
    r=new shape(sh_polystroke);
    for(se=s1->strokes;se;se=se->next)
    {
      r->add_se(se->x,se->y,se->r+s2->r,se->arg1,se->arg2);
      if (se->next==s1->strokes) break;
    }
    return r;
  }

  if (k1==sh_circl && k2==sh_polystroke)
  {
    r=new shape(sh_polystroke);
    for(se=s2->strokes;se;se=se->next)
    {
      r->add_se(se->x,se->y,se->r+s1->r,opp_arg(se->arg1),opp_arg(se->arg2));
      if (se->next==s2->strokes) break;
    }
    return r;
  }

  if (k1==sh_polystroke || k2==sh_polystroke)
    return combine_polystrokes(s1,s2);

  r=new shape(0);
  r->r =s1->r +s2->r;
  r->lr=s1->lr+s2->lr;
  r->ud=s1->ud+s2->ud;

  if (r->r)
    if (r->lr || r->ud) r->kind=sh_roundrect;
    else r->kind=sh_circl;
  else if (r->ud) r->kind=sh_rectangle;
  else if (r->lr) r->kind=sh_hlin;
  else r->kind=sh_point;
  r->calc_maxima();
  return r;
}
